from typing import Iterable
from django.contrib.auth.decorators import login_required, permission_required
from django.http.response import HttpResponse
from esi.decorators import token_required
from django.shortcuts import render, redirect
from django.db.models import QuerySet, Avg, Max, Min, Sum
from eveuniverse.models import EveEntity, EveMarketGroup, EveRegion, EveType
from marketmanager.models import Order, Structure
import datetime
import json

from allianceauth.services.hooks import get_extension_logger

logger = get_extension_logger(__name__)

CHARACTER_SCOPES = [
    'esi-markets.read_character_orders.v1',
    'esi-markets.structure_markets.v1',
    'esi-universe.read_structures.v1',
]

CORPORATION_SCOPES = [
    'esi-markets.read_corporation_orders.v1',
    'esi-markets.structure_markets.v1',
    'esi-characters.read_corporation_roles.v1',
    'esi-corporations.read_structures.v1',
    'esi-universe.read_structures.v1',
]


@login_required
@permission_required("marketmanager.view_marketbrowser")
def marketbrowser(request):
    region_id = request.GET.get('region_id', None)
    type_id = request.GET.get('type_id', None)
    all_regions = EveRegion.objects.all()
    all_market_groups = EveMarketGroup.objects.all()

    try:
        eve_type, eve_type_fetched = EveType.objects.get_or_create_esi(id=type_id)
    except Exception:
        eve_type = None

    try:
        eve_region, everegion_fetched = EveRegion.objects.get_or_create_esi(
            id=region_id)
    except Exception:
        eve_region = None

    # Sell Orders
    if eve_region is None:
        sell_orders = Order.objects.filter(
            eve_type=type_id,
            is_buy_order=False
        ).select_related('eve_region')

    else:
        sell_orders = Order.objects.filter(
            eve_type=type_id,
            eve_region=region_id,
            is_buy_order=False
        ).select_related('eve_region')

    sell_order_locations = []
    for order in sell_orders:
        sell_order_locations.append(order.location_id)
    eveentities_resolved, structures_resolved = bulk_location_resolver(
        sell_order_locations)

    for order in sell_orders:
        if eveentities_resolved.to_name(order.location_id) != "":
            order.location_resolved = eveentities_resolved.to_name(order.location_id)
        elif structures_resolved[order.location_id].name != "":
            order.location_resolved = structures_resolved[order.location_id].name
        else:
            order.location_resolved = order.location_id

        order.expiry_calculated = order.issued + datetime.timedelta(days=order.duration)

    sell_order_stats = order_stats(sell_orders)

    # Buy Orders
    if eve_region is None:
        buy_orders = Order.objects.filter(
            eve_type=type_id,
            is_buy_order=True
        ).select_related('eve_region')

    else:
        buy_orders = Order.objects.filter(
            eve_type=type_id,
            eve_region=region_id,
            is_buy_order=True
        ).select_related('eve_region')

    buy_order_locations = []
    for order in buy_orders:
        buy_order_locations.append(order.location_id)
    eveentities_resolved, structures_resolved = bulk_location_resolver(
        buy_order_locations)

    for order in buy_orders:
        if eveentities_resolved.to_name(order.location_id) != "":
            order.location_resolved = eveentities_resolved.to_name(order.location_id)
        else:
            try:
                order.location_resolved = structures_resolved[order.location_id].name
            except KeyError:
                order.location_resolved = order.location_id

        order.expiry_calculated = order.issued + datetime.timedelta(days=order.duration)

    buy_order_stats = order_stats(buy_orders)

    render_items = {
        "all_regions": all_regions,
        "all_market_groups": all_market_groups,
        "eve_region": eve_region,
        "eve_type": eve_type,
        "sell_orders": sell_orders,
        "buy_orders": buy_orders,
        "sell_order_stats": sell_order_stats,
        "buy_order_stats": buy_order_stats
    }
    return render(request, "marketmanager/marketbrowser.html", render_items)


@login_required
@permission_required("marketmanager.view_marketbrowser")
def marketbrowser_autocomplete(request):
    if request.is_ajax():
        search_query = request.GET.get('term')

    autocomplete_query = EveType.objects.filter(
        name__istartswith=search_query,
        eve_market_group__isnull=False
    )
    result = []
    for possible in autocomplete_query:
        data = {}
        data['label'] = possible.name
        data['value'] = possible.id
        result.append(data)
    dump = json.dumps(result)

    mimetype = 'application/json'
    return HttpResponse(dump, mimetype)


@login_required
@permission_required("marketmanager.view_marketbrowser")
def item_selector(request):
    data = EveMarketGroup.objects.all()
    return render(request, "marketmanager/item_selector.html", data)


@login_required
@token_required(scopes=CHARACTER_SCOPES)
def add_char(request, token):
    return redirect('marketmanager:marketbrowser')


@login_required
@token_required(scopes=CORPORATION_SCOPES)
def add_corp(request, token):
    return redirect('marketmanager:marketbrowser')


def location_resolver(location_id) -> str:
    if location_id >= 60000000 and location_id <= 64000000:
        # EveStation (Range: 60000000 - 64000000)
        # EveEntity has its own resolver
        # but i dont want Structures to slip through
        # and spam ESI errors
        return EveEntity.objects.resolve_name(location_id)
    else:
        try:
            return Structure.objects.get(structure_id=location_id).name
        except Exception as e:
            logger.error(e)
            return str(location_id)


def bulk_location_resolver(location_ids: Iterable[int]):
    bulk_eve_entity_ids = []
    bulk_structure_ids = []

    for location_id in location_ids:
        if location_id >= 60000000 and location_id <= 64000000:
            # EveStation (Range: 60000000 - 64000000)
            # EveEntity has its own resolver
            # but i dont want Structures to slip through
            # and spam ESI errors
            bulk_eve_entity_ids.append(location_id)
        else:
            bulk_structure_ids.append(location_id)

    eveentity_resolver = EveEntity.objects.bulk_resolve_names(bulk_eve_entity_ids)
    structure_resolver = Structure.objects.in_bulk(bulk_structure_ids)

    return eveentity_resolver, structure_resolver


def order_stats(orders):
    # Returns a specific set of stats for the item_details template
    fifth_percentile = "Fifth Percentile FAKE"
    weighted_average = "Weighted Average Fake"
    median = "Median Fake"
    volume = orders.aggregate(volume=Sum('volume_remain'))["volume"]
    return {
        "fifth_percentile": fifth_percentile,
        "weighted_average": weighted_average,
        "median": median,
        "volume": volume}



from allianceauth.eveonline.models import EveCharacter
from esi.models import Token
from .providers import get_characters_character_id_roles_from_token
import random


def get_corp_token(corporation_id: int, scopes: list, req_roles: list) -> Token:
    """
    Helper method to get a token for a specific character from a specific corp with specific scopes
    :param corp_id: Corp to filter on.
    :param scopes: array of ESI scope strings to search for.
    :param req_roles: roles required on the character.
    :return: :class:esi.models.Token or False
    """
    if 'esi-characters.read_corporation_roles.v1' not in scopes:
        scopes.append("esi-characters.read_corporation_roles.v1")

    char_ids = EveCharacter.objects.filter(
        corporation_id=corporation_id).values('character_id')
    tokens = Token.objects \
        .filter(character_id__in=char_ids) \
        .require_scopes(scopes)

    for token in tokens:
        roles = get_characters_character_id_roles_from_token(token)
        has_roles = False
        for role in roles.get('roles', []):
            if role in req_roles:
                has_roles = True

        if has_roles:
            return token
        else:
            pass  # TODO Maybe remove token?

    return False


def get_random_market_token() -> Token:
    """Very specific edge case, we need _any_ token in order to view data on public structures.

    Args:
        scopes: array of ESI scope strings to search for.

    Returns:
        Matching token
    """
    required_scopes = ['esi-markets.structure_markets.v1']
    random_token = random.choice(
        Token.objects.all().require_scopes(
            required_scopes
            )
        )
    return random_token



# from django.contrib import admin

# Register your models for the admin site here.

from django.contrib import admin

from marketmanager.models import Config, Order, Structure, Webhook, WatchConfig

from allianceauth.services.hooks import get_extension_logger

logger = get_extension_logger(__name__)

@admin.register(Config)
class ConfigAdmin(admin.ModelAdmin):
    pass

@admin.register(Order)
class OrderAdmin(admin.ModelAdmin):
    pass

@admin.register(Structure)
class StructureAdmin(admin.ModelAdmin):
    pass

@admin.register(Webhook)
class WebhookAdmin(admin.ModelAdmin):
    pass

@admin.register(WatchConfig)
class WatchConfigAdmin(admin.ModelAdmin):
    pass



from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from celery import shared_task

from allianceauth.services.hooks import get_extension_logger
from discord.embeds import Embed
from django.apps import apps
from django.core.exceptions import ObjectDoesNotExist
from esi.models import Token

from eveuniverse.models import (
    EveSolarSystem,
    EveRegion,
    EveType,
)

from marketmanager.models import WatchConfig, Config, Order, Structure

from marketmanager.providers import (
    get_markets_region_id_orders,
    get_characters_character_id_orders,
    get_corporations_corporation_id_orders,
    get_universe_structures,
    get_universe_structures_structure_id,
    get_corporations_corporation_id_structures
)
from marketmanager.task_helpers import get_corp_token, get_random_market_token

logger = get_extension_logger(__name__)


def discord_bot_active() -> bool:
    return apps.is_installed("aadiscordbot")


if discord_bot_active():
    import aadiscordbot.tasks


@shared_task
def fetch_public_market_orders():
    """Fetch&Save Public Market Orders for configured regions
    bulk calls fetch_markets_region_id_orders(region_id: int)"""
    for region in Config.objects.get(id=1).fetch_regions.all():
        fetch_markets_region_id_orders.delay(region.id)


@shared_task
def fetch_markets_region_id_orders(region_id: int):
    for order in get_markets_region_id_orders(region_id):

        order_eve_type, order_eve_type_fetched = EveType.objects.get_or_create_esi(
            id=order["type_id"],
            enabled_sections=[EveType.Section.MARKET_GROUPS]
        )
        order_eve_solar_system, order_eve_solar_system_fetched = EveSolarSystem.objects.get_or_create_esi(
            id=order["system_id"]
        )

        Order.objects.update_or_create(
            order_id=order["order_id"],
            defaults={
                'eve_type': order_eve_type,
                'duration': order["duration"],
                'is_buy_order': order["is_buy_order"],
                'issued': order["issued"],
                'location_id': order["location_id"],
                'min_volume': order["min_volume"],
                'price': order["price"],
                'range': order["range"],
                'eve_solar_system': order_eve_solar_system,
                'eve_region': order_eve_solar_system.eve_constellation.eve_region,
                'volume_remain': order["volume_remain"],
                'volume_total': order["volume_total"]},
        )


@shared_task
def fetch_all_character_orders():
    """Fetch&Save every Characters Market Orders
    bulk calls fetch_characters_character_id_orders(character_id)"""

    character_ids = Token.objects.values_list('character_id').require_scopes(
        ["esi-markets.read_character_orders.v1"])
    unique_character_ids = list(dict.fromkeys(character_ids))
    for character_id in unique_character_ids:
        fetch_characters_character_id_orders.delay(character_id[0])


@shared_task
def fetch_characters_character_id_orders(character_id: int):
    """Fetch&Save a single Characters Market Orders
    bulk called by fetch_all_character_orders()

    Parameters
    ----------
    corporation_id: int
        Should match a valid Character ID"""
    for order in get_characters_character_id_orders(character_id):

        order_eve_type, order_eve_type_fetched = EveType.objects.get_or_create_esi(
            id=order["type_id"],
            enabled_sections=[EveType.Section.MARKET_GROUPS]
        )
        order_eve_region, order_eve_region_fetched = EveRegion.objects.get_or_create_esi(
            id=order["region_id"]
        )
        if order["is_buy_order"] is None:
            order_is_buy_order = False
        else:
            order_is_buy_order = True

        Order.objects.update_or_create(
            order_id=order["order_id"],
            defaults={
                'eve_type': order_eve_type,
                'duration': order["duration"],
                'is_buy_order': order_is_buy_order,
                'is_corporation': order["is_corporation"],
                'issued': order["issued"],
                'location_id': order["location_id"],
                'eve_region': order_eve_region,
                'min_volume': order["min_volume"],
                'price': order["price"],
                'escrow': order["escrow"],
                'range': order["range"],
                'volume_remain': order["volume_remain"],
                'volume_total': order["volume_total"]},
        )


@shared_task
def fetch_all_corporation_orders():
    """Fetch&Save every Corporations Market Orders
    bulk calls fetch_corporations_corporation_id_orders(corporation_id)"""
    for corporation in EveCorporationInfo.objects.all():
        fetch_corporations_corporation_id_orders.delay(corporation.corporation_id)


@shared_task
def fetch_corporations_corporation_id_orders(corporation_id: int):
    """Fetch&Save a Corporations Market Orders
    Is Bulk-Called by fetch_all_corporation_orders()

    Parameters
    ----------
    corporation_id: int
        Should match a valid Corporation ID"""
    scopes = ["esi-markets.read_corporation_orders.v1"]
    req_roles = ["Accountant", "Trader"]

    token = get_corp_token(corporation_id, scopes, req_roles)
    if token is False:
        logger.error(f"No Token for Corporation {corporation_id}")
        return

    for order in get_corporations_corporation_id_orders(corporation_id, token):

        order_eve_type, order_eve_type_fetched = EveType.objects.get_or_create_esi(
            id=order["type_id"],
            enabled_sections=[EveType.Section.MARKET_GROUPS]
        )
        order_eve_corporation = EveCorporationInfo.objects.get(
            corporation_id=corporation_id
        )
        order_eve_region, order_eve_region_fetched = EveRegion.objects.get_or_create_esi(
            id=order["region_id"]
        )
        try:
            order_eve_character = EveCharacter.objects.get(
                character_id=order["issued_by"]
            )
        except ObjectDoesNotExist:
            EveCharacter.objects.create_character(order["issued_by"])
            order_eve_character = EveCharacter.objects.get(
                character_id=order["issued_by"]
            )

        if order["is_buy_order"] is None:
            order_is_buy_order = False
        else:
            order_is_buy_order = True

        Order.objects.update_or_create(
            order_id=order["order_id"],
            defaults={
                'eve_type': order_eve_type,
                'duration': order["duration"],
                'is_buy_order': order_is_buy_order,
                'is_corporation': True,
                'issued': order["issued"],
                'issued_by_character': order_eve_character,
                'issued_by_corporation': order_eve_corporation,
                'wallet_division': order["wallet_division"],
                'location_id': order["location_id"],
                'eve_region': order_eve_region,
                'min_volume': order["min_volume"],
                'price': order["price"],
                'escrow': order["escrow"],
                'range': order["range"],
                'volume_remain': order["volume_remain"],
                'volume_total': order["volume_total"]},
        )


@shared_task()
def fetch_public_structures():
    for structure_id in get_universe_structures(filter="market"):
        structure = get_universe_structures_structure_id(
            structure_id, get_random_market_token())

        structure_eve_solar_system, structure_eve_solar_system_fetched = EveSolarSystem.objects.get_or_create_esi(
            id=structure["solar_system_id"]
        )
        structure_eve_type, structure_eve_type_fetched = EveType.objects.get_or_create_esi(
            id=structure["type_id"],
            enabled_sections=[EveType.Section.MARKET_GROUPS]
        )

        Structure.objects.update_or_create(
            structure_id=structure_id,
            defaults={
                'name': structure["name"],
                'owner_id': structure["owner_id"],
                'solar_system': structure_eve_solar_system,
                'eve_type': structure_eve_type
            }
        )


@shared_task()
def fetch_all_corporations_structures():
    for corporation in EveCorporationInfo.objects.all():
        fetch_corporations_corporation_id_structures.delay(corporation.corporation_id)


@shared_task()
def fetch_corporations_corporation_id_structures(corporation_id: int):
    scopes = ["esi-corporations.read_structures.v1"]
    req_roles = ["Station_Manager"]

    token = get_corp_token(corporation_id, scopes, req_roles)
    if token is False:
        logger.error(f"No Token for Corporation {corporation_id}")
        return

    for structure in get_corporations_corporation_id_structures(corporation_id, token):
        if structure["services"]["market"] == "Offline" or "Online":
            structure_eve_solar_system, structure_eve_solar_system_fetched = EveSolarSystem.objects.get_or_create_esi(
                id=structure["solar_system_id"]
            )
            structure_eve_type, structure_eve_type_fetched = EveType.objects.get_or_create_esi(
                id=structure["type_id"],
                enabled_sections=[EveType.Section.MARKET_GROUPS]
            )
            Structure.objects.update_or_create(
                structure_id=structure["structure_id"],
                defaults={
                    'name': structure["name"],
                    'owner_id': structure["corporation_id"],
                    'solar_system': structure_eve_solar_system,
                    'eve_type': structure_eve_type
                }
            )


# @shared_task
# def run_market_checks():
#     # filter by unique typeids in Watchconfig
#     for type in WatchConfig.objects.values("Type").distinct():
#         # for each typeID, process its WatchConfig

#         embed = template_discord_embed(type)
#         # idk do stuff

#         #

#         # Comparators

#         # Volume
#         embed.add_field("Volume Insufficient",
#                         f"{config.locations}: {current_quant}/{config.volume}")
#         # Price
#         embed.add_field("Overpriced Order",
#                         f"{config.locations}: {current_quant}/{config.volume}")
#         # Jita %
#     return True


# def template_discord_embed(evetype: EveType) -> Embed:
#     """Prep a standard Embed template, WatchConfigs may further modify/add fields"""
#     embed = Embed

#     embed.title = f"AA Market Manager: {evetype.name}"
#     embed.thumbnail = evetype.icon_url
#     embed.color = "Red"
#     embed.description = "The Following Rules were Failed"

#     return embed



from django.urls import path, re_path

from . import views

app_name = "marketmanager"

urlpatterns = [
    path('', views.marketbrowser, name="index"),
    path('marketbrowser', views.marketbrowser, name="marketbrowser"),
    path('marketmanager', views.marketbrowser, name="marketmanager"),
    path('marketbrowser/ajax/search',  views.marketbrowser_autocomplete, name="marketbrowser_autocomplete"),
    path('char/add', views.add_char, name="add_char"),
    path('corp/add', views.add_corp, name="add_corp")
]



from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from django.db import models
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _

from eveuniverse.models import (
    EveSolarSystem,
    EveConstellation,
    EveRegion,
    EveType,
)


class General(models.Model):
    """Meta model for app permissions"""

    class Meta:
        managed = False
        default_permissions = ()
        permissions = (("basic_access", "Can access this app"),)


class Config(models.Model):
    fetch_regions = models.ManyToManyField(EveRegion, verbose_name=_("Fetch Regions"))

    def save(self, *args, **kwargs):
        if not self.pk and Config.objects.exists():
            # Force a single object
            raise ValidationError('There is can be only one \
                                    AnalyticsIdentifier instance')
        self.pk = self.id = 1  # If this happens to be deleted and recreated, force it to be 1
        return super().save(*args, **kwargs)


class Order(models.Model):
    """An EVE Market Order"""
    order_id = models.PositiveBigIntegerField(
        _("Order ID"),
        help_text="Unique order ID",
        primary_key=True)
    eve_type = models.ForeignKey(
        EveType,
        verbose_name=_("Type"),
        on_delete=models.CASCADE)
    duration = models.PositiveIntegerField(
        _("Duration"),
        help_text="Number of days the order was valid for (starting from the issued date). An order expires at time issued + duration")
    is_buy_order = models.BooleanField(
        _("Buy Order"),
        default=False,
        help_text="True if the order is a bid (buy) order")
    issued = models.DateTimeField(
        _("Issued"),
        help_text="Date and time when this order was issued",
        auto_now=False,
        auto_now_add=False)
    location_id = models.PositiveBigIntegerField(
        _("Location ID"),
        help_text="ID of the location where order was placed")
    eve_solar_system = models.ForeignKey(
        EveSolarSystem,
        verbose_name=_("System"),
        on_delete=models.CASCADE,
        null=True,
        blank=True,)
    eve_region = models.ForeignKey(
        EveRegion,
        verbose_name=_("Region"),
        on_delete=models.CASCADE)
    min_volume = models.PositiveIntegerField(
        _("Minimum Volume"),
        null=True,
        blank=True,
        help_text="For buy orders, the minimum quantity that will be accepted in a matching sell order")
    price = models.DecimalField(
        _("Price"),
        max_digits=20,
        decimal_places=2,
        help_text="Cost per unit for this order")
    escrow = models.DecimalField(
        _("Escrow"),
        max_digits=20,
        decimal_places=2,
        null=True,
        blank=True,
        help_text="For buy orders, the amount of ISK in escrow")
    RANGE_CHOICES = [
        ('1', '1'), ('10', '10'), ('2', '2'), ('20', '20'), ('3', '3'),
        ('30', '30'), ('4', '4'), ('40', '40'), ('5', '5'),
        ('region', _('Region')),
        ('solarsystem', _('Solar System')),
        ('station', _('Station'))]
    range = models.CharField(
        _("Order Range"),
        max_length=20,
        choices=RANGE_CHOICES,
        help_text="Valid order range, numbers are ranges in jumps")
    volume_remain = models.PositiveIntegerField(
        _("Volume Remaining"),
        help_text="Quantity of items still required or offered")
    volume_total = models.PositiveIntegerField(
        _("Volume Total"),
        help_text="Quantity of items required or offered at time order was placed")
    is_corporation = models.BooleanField(
        _("Is Corporation"),
        default=False,
        help_text="Signifies whether the buy/sell order was placed on behalf of a corporation.")
    wallet_division = models.PositiveIntegerField(
        _("Wallet Division"),
        null=True,
        blank=True,
        help_text="The corporation wallet division used for this order.")
    issued_by_character = models.ForeignKey(
        EveCharacter,
        verbose_name=_("Character"),
        on_delete=models.SET_NULL,
        null=True,
        blank=True)
    issued_by_corporation = models.ForeignKey(
        EveCorporationInfo,
        verbose_name=_("Corporation"),
        on_delete=models.SET_NULL,
        null=True,
        blank=True,)
    STATE_CHOICES = [
        ('', ''),
        ('cancelled', _('Cancelled')),
        ('expired ', _('Expired'))]
    state = models.CharField(
        _("Order State"),
        max_length=20,
        choices=RANGE_CHOICES,
        help_text="Current order state, Only valid for Authenticated order History. Will not update from Public Market Data.")
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)


class Structure(models.Model):
    """An EVE Online Upwell Structure"""
    structure_id = models.PositiveBigIntegerField(
        _("Structure ID"),
        primary_key=True)
    name = models.CharField(
        _("Name"),
        max_length=100)
    owner_id = models.IntegerField(_("Owner ID"))
    solar_system = models.ForeignKey(
        EveSolarSystem,
        verbose_name=_("Solar System"),
        on_delete=models.CASCADE)
    eve_type = models.ForeignKey(
        EveType,
        verbose_name=_(""),
        on_delete=models.CASCADE)
    pull_market = models.BooleanField(
        _("Pull Market Orders"),
        help_text="Useful to ignore specific structures for _reasons_",
        default=True)


class Webhook(models.Model):
    name = models.CharField(
        _("Name"),
        max_length=100)
    webhook = models.URLField(_("URL"), max_length=200)


class WatchConfig(models.Model):
    """Rules to Watch"""
    # Item Rules, Run once for each Item
    buy_order = models.BooleanField(_("Buy Order"))
    type = models.ManyToManyField(EveType, verbose_name=_("EVE Types"))
    # Location Rules, Combined
    location = models.ManyToManyField(Structure, verbose_name=_(""))
    solar_system = models.ManyToManyField(EveSolarSystem, verbose_name=_(""))
    constellation = models.ManyToManyField(EveConstellation, verbose_name=_(""))
    region = models.ManyToManyField(EveRegion, verbose_name=_(""))
    # Filter
    structure_type = models.ManyToManyField(EveType,
                                            help_text="Filter by structure Type/Size/Docking (ie, forts/keeps for cap fuel)",
                                            verbose_name=_("Structure Type Filter"),
                                            related_name="structure_type")
    # Comparators
    volume = models.IntegerField(_("Volume"),
                                 help_text="Set to Zero to check ANY/EVERY order against Price")
    price = models.IntegerField(_("Price"),
                                help_text="Set to Zero to check ANY/EVERY order against Volume")
    jita_compare_percent = models.IntegerField(_("Jita Comparison %"),
                                               help_text="If set ignores Flat price value")

    # Destinations

    webhooks = models.ManyToManyField(Webhook, verbose_name=_("Webhooks"))
